#ifndef PT_H
#define PT_H

#include <cmath>
#include <iostream>
#include "OutilsMaths.h"

struct pt
{
   float x, y, z;
   pt () : x(0.0f), y(0.0f), z(0.0f) { }
   pt (const float _x, const float _y, const float _z) : x(_x), y(_y), z(_z) { }
   pt (const pt &p) : x(p.x), y(p.y), z(p.z) { }
   pt operator+ (const pt &p) const
   {
      return pt (x+p.x, y+p.y, z+p.z);
   }
   pt operator- (const pt &p) const
   {
      return pt (x-p.x, y-p.y, z-p.z);
   }
   pt operator- () const
   {
      return pt (-x, -y, -z);
   }
   pt operator* (const float f) const
   {
      return pt (x * f, y * f, z * f);
   }
   float lg () const
   {
      return std::sqrt (x*x+y*y+z*z);
   }
   //
   // Produit scalaire (U.V). Intressant: (U.V)==|U||V|cos(theta)
   //
   float ProduitScalaire (const pt &p) const
   {
      return x*p.x+y*p.y+z*p.z;
   }
   //
   // Dterminant d'une matrice 2D contenant
   // +-       -+
   // |  x   y  |
   // | p.x p.y |
   // +-       -+
   //
   // Le cas 3D poserait problme si z et p.z valaient zro tous les deux
   //
   // a11a22a33 + a12a23a31 + a13a21a32 - a11a23a32 - a12a21a33 - a13a22a31
   //
   float CrossProduct (const pt &p) const // retoucher pour le 3D
   {
      return x*p.y-y*p.x;
   }
   //
   // p. 293, intro  l'algbre linaire
   //
   pt ProduitVectoriel (const pt &p) const
   {
      return pt (1.0f, 0.0f, 0.0f) * (y * p.z - z * p.y) -
             pt (0.0f, 1.0f, 0.0f) * (z * p.x - x * p.z) +
             pt (0.0f, 0.0f, 1.0f) * (x * p.y - y * p.x);
   }

   pt normaliser () const
   {
      float _lg = lg ();
      return pt (x/_lg,y/_lg,z/_lg);
   }
   bool operator > (const pt &p) const
   {
      if (y > p.y + OutilsMaths::TOLERANCE) return true;
      if (y < p.y - OutilsMaths::TOLERANCE) return false;
      return x > p.x;
   }
   bool operator >= (const pt &p) const
   {
      if (y > p.y + OutilsMaths::TOLERANCE) return true;
      if (y < p.y - OutilsMaths::TOLERANCE) return false;
      return x >= p.x;
   }
   bool operator < (const pt &p) const
   {
      if (y < p.y - OutilsMaths::TOLERANCE) return true;
      if (y > p.y + OutilsMaths::TOLERANCE) return false;
      return x < p.x;
   }
   bool operator <= (const pt &p) const
   {
      if (y < p.y - OutilsMaths::TOLERANCE) return true;
      if (y > p.y + OutilsMaths::TOLERANCE) return false;
      return x <= p.x;
   }
   bool operator == (const pt &p) const
   {
      return OutilsMaths::Egaux (x, p.x) &&
             OutilsMaths::Egaux (y, p.y) &&
             OutilsMaths::Egaux (z, p.z);
   }
   const pt & operator = (const pt &p)
   {
      x = p.x;
      y = p.y;
      z = p.z;
      return *this;
   }
};

std::ostream &operator<< (std::ostream &o, const pt &p);

//
// Exceptions possibles
//
class VecteurLongueurNulle { };

//
// Calcule et retourne l'angle (en radians) entre deux vecteurs.
// Note: les deux vecteurs doivent tre non nuls
//
float CalculerAngleEntreVecteurs (const pt &, const pt &) throw (VecteurLongueurNulle);

//
// Calcule et retourne l'angle (en radians) ABC. Les vecteurs AB et BC doivent tre non nuls
//
float CalculerAngle (const pt &, const pt &, const pt &) throw (VecteurLongueurNulle);

//
// Retourne vrai ssi le vecteur p0p1 est accessible du vecteur
// p0p2 par un virage en sens horaire
//
bool VirageHoraire (const pt &, const pt &, const pt &);
//
// Retourne vrai ssi le vecteur p0p1 est accessible du vecteur
// p0p2 par un virage en sens antihoraire
//
bool VirageAntihoraire (const pt &, const pt &, const pt &);
//
// Retourne vrai ssi le vecteur p0p1 est  gauche du vecteur p0p2
//
bool AGauche (const pt &, const pt &, const pt &);
//
// Retourne vrai ssi le vecteur p0p1 est  droite du vecteur p0p2
//
bool ADroite (const pt &, const pt &, const pt &);

//
// Clockwise?
//
// v(p0p1) clockwise from v(p0p2) iff (p1-p0)*(p2-p0)>0
// --> Permet aussi de savoir si p0 est  gauche de p1p2
//     en examinant si v(p0p2) est clockwise ou non de v(p0p1)
//

#endif // !defined PT_H
